/****************************************************************************

X Play Gizmo 1.1    (for use with NCSA Mosaic for the X Window System)
Alan M. Braverman                                              10/3/93 
=============================================================================

-----------------------------------------------------------------------------
COMPILING: (xplaygizmo requires Motif 1.1 to compile)

SUN:  make -f makefile.sun
SGI:  make -f makefile.sgi
-----------------------------------------------------------------------------


usage: xplaygizmo [-p] [-q]  player-command  data-file

This utility was written for use with NCSA Mosaic for the X Window
System.  X Play Gizmo causes a small control panel to pop up after
Mosaic invokes an external sound player (e.g. showaudio) or movie
viewer (e.g. mpeg_play).  After installing X Play Gizmo, sound and movie
files may be played multiple times, and saved locally, without multiple
data transfers.  The control panel contains the following buttons:

  PLAY invokes the sound player or movie viewer.  Pressing the button
  during a sound or movie causes a restart.

  STOP kills the process.

  SAVE allows the sound or movie file to be saved to a specified
  filename.

  EXIT quits X Play Gizmo.

  PAUSE (new for version 1.1) will pause a sound ???


X Play Gizmo also recognizes two command-line flags:

  -p  causes X Play Gizmo to automatically play the request upon
      execution, without waiting for the PLAY button to be pressed.

  -q  causes X Play Gizmo to automatically quit once the request has
      completed playing.  Pressing  STOP will not cause a quit,
      however; the sound or movie must reach the end.

=============================================================================

Mosaic 1.2 users:

To use X Play Gizmo with NCSA Mosaic 1.2, add the following lines to
your .Xdefaults file, or replace showaudio and mpeg_play with your own
sound and movie players.  

  XMosaic*audioPlayerCommand: xplaygizmo showaudio
  XMosaic*aiffPlayerCommand:  xplaygizmo sfplay
  XMosaic*mpegViewerCommand:  xplaygizmo mpeg_play

Either or both of the command-line options, -p and -q, may be added as
follows: 

  XMosaic*audioPlayerCommand: xplaygizmo -p -q showaudio
  XMosaic*mpegViewerCommand:  xplaygizmo -p mpeg_play

or 
 
  XMosaic*audioPlayerCommand: xplaygizmo -p -q /usr/local/bin/showaudio

if necessary.

If you do not have a .Xdefaults file, you may create one with no other
text than this.  For more information on the X resources used by NCSA
Mosaic, open the following URL with the OPEN feature:

  http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/resources.html

=============================================================================

Mosaic 2.0 users:

To use X Play Gizmo with NCSA Mosaic 2.0 (prerelease 4 and later), add
the following lines to your .mailcap (or equivalent) file, or create a
.mailcap file with no other text than:

  audio/*; xplaygizmo showaudio %s
  video/mpeg; xplaygizmo mpeg_play %s

Either or both of the command-line options, -p and -q, may be added as
follows: 

  audio/*; xplaygizmo -p -q showaudio %s
  video/mpeg; xplaygizmo -p mpeg_play %s

or 
 
  audio/*; xplaygizmo -p -q /usr/local/bin/showaudio %s

if necessary.

For more information on using mailcap files to map MIME types to
external viewers, see Help on Version 2.0 from within Mosaic:

http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/help-on-version-2.0.html

--
Alan Braverman
Software Development Group
National Center for Supercomputing Applications
alanb@ncsa.uiuc.edu

*****************************************************************************/

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <sys/wait.h>
#include "xplaygizmo.h"

#if defined(__sgi)
#include <malloc.h>
#endif

char  *buttons[] = {"PLAY", "STOP", "QUIT", "SAVE", "PAUSE"};
char  play_option=0, quit_option=0, user_stop=0;
int   process_id=0, save_up=0, pause_mode=0;
char  *temp_file_name;
extern char *color_resources[];
Widget toplevel;

#define play_width 32
#define play_height 32
static char play_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0xf8, 0x00, 0x00, 0x00, 0xf8, 0x01, 0x00, 0x00, 0xf8, 0x03, 0x00,
   0x00, 0xb8, 0x07, 0x00, 0x00, 0x38, 0x0f, 0x00, 0x00, 0x38, 0x1e, 0x00,
   0x00, 0x38, 0x3c, 0x00, 0x00, 0x38, 0x78, 0x00, 0x00, 0x38, 0x3c, 0x00,
   0x00, 0x38, 0x1e, 0x00, 0x00, 0x38, 0x0f, 0x00, 0x00, 0xb8, 0x07, 0x00,
   0x00, 0xf8, 0x03, 0x00, 0x00, 0xf8, 0x01, 0x00, 0x00, 0xf8, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x3e, 0x02, 0x3c, 0x41, 0x42, 0x02, 0x42, 0x41, 0x42, 0x02, 0x42, 0x22,
   0x42, 0x02, 0x42, 0x22, 0x42, 0x02, 0x42, 0x14, 0x3e, 0x02, 0x7e, 0x08,
   0x02, 0x02, 0x42, 0x08, 0x02, 0x02, 0x42, 0x08, 0x02, 0x02, 0x42, 0x08,
   0x02, 0x7e, 0x42, 0x08, 0x00, 0x00, 0x00, 0x00};

#define quit_width 32
#define quit_height 32
static char quit_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x84, 0x7c, 0x3e, 0x04, 0x84, 0x10, 0x08, 0x04, 0x48, 0x10, 0x08,
   0x04, 0x48, 0x10, 0x08, 0x3c, 0x30, 0x10, 0x08, 0x04, 0x30, 0x10, 0x08,
   0x04, 0x48, 0x10, 0x08, 0x04, 0x48, 0x10, 0x08, 0x04, 0x84, 0x10, 0x08,
   0xfc, 0x84, 0x7c, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00,
   0x00, 0x00, 0xf0, 0x03, 0x00, 0x00, 0xe0, 0x0f, 0xfc, 0xff, 0xff, 0x3f,
   0xfc, 0xff, 0xff, 0x3f, 0x00, 0x00, 0xe0, 0x0f, 0x00, 0x00, 0xf0, 0x03,
   0x00, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

#define stop_width 32
#define stop_height 32
static char stop_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0xf8, 0x7f, 0x00, 0x00, 0xf8, 0x7f, 0x00, 0x00, 0xf8, 0x7f, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0xf8, 0x7f, 0x00, 0x00, 0xf8, 0x7f, 0x00, 0x00, 0xf8, 0x7f, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x3c, 0xfe, 0x3c, 0x3e, 0x42, 0x10, 0x42, 0x42, 0x02, 0x10, 0x42, 0x42,
   0x02, 0x10, 0x42, 0x42, 0x0c, 0x10, 0x42, 0x42, 0x30, 0x10, 0x42, 0x3e,
   0x40, 0x10, 0x42, 0x02, 0x40, 0x10, 0x42, 0x02, 0x42, 0x10, 0x42, 0x02,
   0x3c, 0x10, 0x3c, 0x02, 0x00, 0x00, 0x00, 0x00};

#define save_width 32
#define save_height 32
static char save_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7f, 0x00, 0x00,
   0x80, 0x80, 0x00, 0x00, 0x40, 0x00, 0x01, 0x00, 0xf0, 0xff, 0xff, 0x0f,
   0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08,
   0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08,
   0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08,
   0x10, 0x00, 0x00, 0x08, 0xf0, 0xff, 0xff, 0x0f, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x3c, 0x3c, 0x82, 0x7e, 0x42, 0x42, 0x82, 0x02, 0x02, 0x42, 0x82, 0x02,
   0x02, 0x42, 0x44, 0x02, 0x0c, 0x42, 0x44, 0x1e, 0x30, 0x7e, 0x44, 0x02,
   0x40, 0x42, 0x28, 0x02, 0x40, 0x42, 0x28, 0x02, 0x42, 0x42, 0x10, 0x02,
   0x3c, 0x42, 0x10, 0x7e, 0x00, 0x00, 0x00, 0x00};

#define pause_width 32
#define pause_height 32
static char pause_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00, 0x00, 0x38, 0x70, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x1e, 0x27, 0x72, 0x3e, 0xa2, 0x28, 0x8a, 0x02, 0xa2, 0x28, 0x0a, 0x02,
   0xa2, 0x28, 0x12, 0x02, 0xa2, 0x28, 0x22, 0x0e, 0x9e, 0x2f, 0x42, 0x02,
   0x82, 0x28, 0x82, 0x02, 0x82, 0x28, 0x82, 0x02, 0x82, 0x28, 0x8a, 0x02,
   0x82, 0xc8, 0x71, 0x3e, 0x00, 0x00, 0x00, 0x00};

/* ---------------------------------------------------------------------- */
void ign() 
{
  int status;
  pid_t pid;

#if defined(sun)
  while( (pid = wait3(&status, WNOHANG, NULL)) > 0) ;
#else 
  while( (pid = waitpid(-1, &status, WNOHANG)) > 0) ;
#endif

  if (pause_mode==0) {            /* if in pause mode, ignore kill signals */
    
    if (quit_option==1)         /* if sound is done playing and quit option */
      if ((user_stop==0) && (save_up==0))        /* and not stopped by user */
	exit(0);                    /* and save dialog is not up, then quit */
    
    if (user_stop==0)           
      process_id=0;
    
    user_stop=0;
  }
  return;
}

/* ----------------------------------------------------------------------- */
void playCallback(Widget w, caddr_t *command_line[], caddr_t call_data)
{
  int    current_pid;

 signal(SIGCHLD,(void (*)())ign);

  if (pause_mode==1)  {    /* if in pause mode, send the kill SIGCONT */
    kill(process_id, SIGCONT);
    pause_mode=0;
    return;
  }

  if (process_id!=0)                   /* kill the currently playing sound */
    {
      user_stop=1; 
      kill(process_id, 9);
      process_id=0;
    }

  current_pid=getpid();                /* get the current process id */
  if ((process_id=fork())==-1)         /* fork off a new process */
    printf("Error -- cannot create new process.\n\n");
  else 
    
    /* only execute the sound player if we are currently in the forked process
       as opposed to having returned to the parent process  */
    
    if (getpid()!=current_pid) {     
      execvp(command_line[0], command_line);  /* execute sound player */
      exit(0);
    }
  return;
}

/* ----------------------------------------------------------------------- */
void stopCallback(Widget w, caddr_t client_data, caddr_t call_data)
{
  if (process_id!=0)
    {
      user_stop=1;
      kill(process_id, 9);
      process_id=0;
      pause_mode=0;
    }
  return;
}

/* ---------------------------------------------------------------------- */
void save_file(Widget w, caddr_t parent_shell_widget, caddr_t call_data)
{
  char *filename, save_command_string[500];
 
  XmSelectionBoxCallbackStruct *scb=(XmSelectionBoxCallbackStruct *)call_data;
  XmStringGetLtoR(scb->value, XmSTRING_DEFAULT_CHARSET, &filename);
  
  save_command_string[0]=NULL;              /* build the string to copy file */
  strcat(save_command_string, "cp ");       /* from tmp dir to selected name */
  strcat(save_command_string, temp_file_name);
  free (temp_file_name);
  strcat(save_command_string, " ");
  strcat(save_command_string, filename);
  system(save_command_string);              /* cp file1 file2 */

  XtPopdown(parent_shell_widget);     /* get rid of the save widget */
  save_up=0;

  return;
}

/* ------------------------------------------------------------------------ */
void cancel_save(Widget w, caddr_t parent_shell_widget, caddr_t call_data)
{
  XtPopdown(parent_shell_widget);    /* get rid of the save widget */
  save_up=0;
  return;
}

/* ---------------------------------------------------------------------- */
void saveCallback(Widget w, caddr_t *sound_file_name, caddr_t call_data)
{
  Widget  saveWidget, parent_shell;
  Arg     wargs[1];
  int     n;

  temp_file_name = strdup ((char *)sound_file_name);
  n=0;

  save_up=1;
  parent_shell=XtCreatePopupShell("Save audio file", 
				  transientShellWidgetClass, 
				  toplevel, wargs, n);
  saveWidget = XtCreateManagedWidget("Save audio file", 
				     xmFileSelectionBoxWidgetClass, 
				     parent_shell, wargs, n);   

  XtAddCallback(saveWidget, XmNokCallback, save_file, parent_shell);
  XtAddCallback(saveWidget, XmNcancelCallback, cancel_save, parent_shell);
  XtUnmanageChild(XmSelectionBoxGetChild(saveWidget, XmDIALOG_HELP_BUTTON));
  XtPopup(parent_shell, XtGrabNone);

  return;
}

/* --------------------------------------------------------------------- */
void quitCallback(Widget w, caddr_t client_data, caddr_t call_data)
{

  /* if the process is still running, we better kill it when we quit,
     otherwise, the sound being played cannot be aborted  */
  
  if (process_id!=0)
    kill(process_id, 9);
  
  XtCloseDisplay(XtDisplay(w));
  exit(0);
  return;
}

/* ------------------------------------------------------------------------ */
void pauseCallback(Widget w, caddr_t client_data, caddr_t call_data)
{
  /* send the process "kill SIGSTOP" to pause the current process  */
  /* if pause_mode==1, then playCallback will send a "kill SIGCONT" */

  if (pause_mode==1)         /* if already in pause mode, do nothing */
    return;
  else
    if (process_id!=0) {
      pause_mode=1;          /* otherwise pause the process with SIGSTOP */
      kill(process_id, SIGSTOP);
    }
  
  return;
}



/* --------------------------------------------------------------------------*/
int handle_options(int argc, char *argv[])
{
  int option, illegal_flag=0; 
  extern int optind;          
  extern char *optarg;        
  
  /* valid arguments:   */
  /* -q     quit after sound plays */
  /* -p     play upon execute      */
  
  while ((option=getopt(argc,argv,"qpQP"))!=EOF)  /* get the arguments */
    switch(option)
      {
      case 'Q':
      case 'q': quit_option=1;
	break;
      case 'P':
      case 'p': play_option=1;
	break;
      default: illegal_flag=1;
      }  /* end of switch */
  if (illegal_flag) ; /* no response to illegal options */
  return optind;
}

/* ------------------------------------------------------------------------ */
main(int argc, char *argv[])
{
  Widget  form, button[5];    /* 0=play, 1=stop, 2=quit, 3=save, 4=pause */
  XtAppContext app;
  Arg     wargs[16];
  Pixmap  play_bitmap, save_bitmap, quit_bitmap, stop_bitmap, pause_bitmap;
  Pixel   foreground, background;
  int     i, n, next_arg;

  static XrmOptionDescRec options[] = {" " };

  next_arg=handle_options(argc,argv);
  
  if (argc<3) {
    printf("\nusage: %s [-p] [-q] <audio player command> <sound file>\n\n",argv[0]);
    exit(1);
  }
  
  toplevel = XtAppInitialize(&app, "XPlayGizmo", options, 0, &argc, argv, color_resources, NULL, 0);

  /* create form widget that manages the buttons */

  form = XtCreateManagedWidget("form", xmFormWidgetClass, toplevel, NULL, 0);

  /* create 5 buttons and assign an XmNactivateCallback to each */
  for(i=0; i<XtNumber(buttons); i++) 
    button[i] = XtCreateWidget(buttons[i], xmPushButtonWidgetClass, form, 
			       NULL, 0); 

  XtAddCallback(button[0], XmNactivateCallback, playCallback, &argv[next_arg]);
  XtAddCallback(button[1], XmNactivateCallback, stopCallback, NULL);
  XtAddCallback(button[2], XmNactivateCallback, quitCallback, NULL);
  XtAddCallback(button[3], XmNactivateCallback, saveCallback, argv[argc-1]);
  XtAddCallback(button[4], XmNactivateCallback, pauseCallback, NULL);

  /* retrieve form widget colors */

  n=0;
  XtSetArg(wargs[n], XmNforeground, &foreground); n++;
  XtSetArg(wargs[n], XmNbackground, &background); n++;
  XtGetValues(form, wargs, n);

  /* create bitmaps for button labels */
  
  play_bitmap=XCreatePixmapFromBitmapData(XtDisplay(button[0]),
    RootWindowOfScreen(XtScreen(button[0])), play_bits, play_width, 
    play_height, foreground, background, 
    DefaultDepthOfScreen(DefaultScreenOfDisplay(XtDisplay(button[0]))));

  save_bitmap=XCreatePixmapFromBitmapData(XtDisplay(button[0]),
    RootWindowOfScreen(XtScreen(button[0])), save_bits, save_width, 
    save_height, foreground, background, 
    DefaultDepthOfScreen(DefaultScreenOfDisplay(XtDisplay(button[0]))));

  stop_bitmap=XCreatePixmapFromBitmapData(XtDisplay(button[0]),
    RootWindowOfScreen(XtScreen(button[0])), stop_bits, stop_width, 
    stop_height, foreground, background, 
    DefaultDepthOfScreen(DefaultScreenOfDisplay(XtDisplay(button[0]))));

  quit_bitmap=XCreatePixmapFromBitmapData(XtDisplay(button[0]),
    RootWindowOfScreen(XtScreen(button[0])), quit_bits, quit_width, 
    quit_height, foreground, background, 
    DefaultDepthOfScreen(DefaultScreenOfDisplay(XtDisplay(button[0]))));

  pause_bitmap=XCreatePixmapFromBitmapData(XtDisplay(button[0]),
    RootWindowOfScreen(XtScreen(button[0])), pause_bits, pause_width, 
    pause_height, foreground, background, 
    DefaultDepthOfScreen(DefaultScreenOfDisplay(XtDisplay(button[0]))));
  
  /* before the buttons are realized, set their attachments */
  /* left, play button = button[0] */
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 0); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 20); n++;
  XtSetArg(wargs[n], XmNlabelType, (XtArgVal)XmPIXMAP); n++;
  XtSetArg(wargs[n], XmNlabelPixmap, (XtArgVal)play_bitmap); n++;
  XtSetValues(button[0], wargs, n);

  
  /* middle left, stop button = button[1] */
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 20); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 40); n++;
  XtSetArg(wargs[n], XmNlabelType, (XtArgVal)XmPIXMAP); n++;
  XtSetArg(wargs[n], XmNlabelPixmap, (XtArgVal)stop_bitmap); n++;
  XtSetValues(button[1], wargs, n);

  /* NEW BUTTON - pause */

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 40); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 60); n++;
  XtSetArg(wargs[n], XmNlabelType, (XtArgVal)XmPIXMAP); n++;
  XtSetArg(wargs[n], XmNlabelPixmap, (XtArgVal)pause_bitmap); n++;
  XtSetValues(button[4], wargs, n);
  
  /* middle right, save button = button[3]  */
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 60); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 80); n++;
  XtSetArg(wargs[n], XmNlabelType, (XtArgVal)XmPIXMAP); n++;
  XtSetArg(wargs[n], XmNlabelPixmap, (XtArgVal)save_bitmap); n++;
  XtSetValues(button[3], wargs, n);
  

  /* right, quit button = button[2]  */
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 80); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 100); n++;
  XtSetArg(wargs[n], XmNlabelType, (XtArgVal)XmPIXMAP); n++;
  XtSetArg(wargs[n], XmNlabelPixmap, (XtArgVal)quit_bitmap); n++;
  XtSetValues(button[2], wargs, n);
  
  /* realize widgets and enter main loop */
  
  XtManageChildren(button, XtNumber(buttons));
  
  XtRealizeWidget(toplevel);
  
  if (play_option==1)
    playCallback(button[0], (caddr_t **) &argv[next_arg], (caddr_t) NULL);

  signal(SIGCHLD,(void (*)())ign);
  
  XtAppMainLoop(app);
}
